/***************************************************************************
 *   Copyright (C) 2007 by David Bitseff                                   *
 *   dbitsef@zipcon.net                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/***************************************************************************
 *   The mutation.h and mutation.c files contain code taken from the flam3 *
 *   application written by Scott Draves.  The functions and variables     *
 *   here have been copied with Scott's permission from the flam3          *
 *   private.h and flam3-genome.c files, and modified to fit this          *
 *   application.                                                          *
 ***************************************************************************/
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>

#include "mutation.h"


void add_to_action(char *action, char *addtoaction) {

   int alen = strlen(action);
   int addlen = strlen(addtoaction);

   if (alen+addlen < flam3_max_action_length)
      strcat(action,addtoaction);
}

void truncate_variations(flam3_genome *g, int max_vars, char *action)
{
   int i, j, nvars, smallest;
   double sv = 0;
   char trunc_note[30];

   for (i = 0; i < g->num_xforms; i++) {
      double d = g->xform[i].density;

      if (d < 0.001 && (g->final_xform_index != i)) {
         sprintf(trunc_note," trunc_density %d",i);
         add_to_action(action,trunc_note);
         flam3_delete_xform(g, i);
      } else {
         do {
            nvars = 0;
            smallest = -1;
            for (j = 0; j < flam3_nvariations; j++) {
               double v = g->xform[i].var[j];
               if (v != 0.0) {
                  nvars++;
                  if (-1 == smallest || fabs(v) < sv) {
                     smallest = j;
                     sv = fabs(v);
                  }
               }
            }
            if (nvars > max_vars) {
               sprintf(trunc_note," trunc %d %d",i,smallest);
               add_to_action(action,trunc_note);
               g->xform[i].var[smallest] = 0.0;
            }
         } while (nvars > max_vars);
      }
   }
}


void test_cp(flam3_genome *cp) {
   cp->time = 0.0;
   cp->interpolation = flam3_interpolation_linear;
   cp->palette_interpolation = flam3_palette_interpolation_hsv;
   cp->background[0] = 0.0;
   cp->background[1] = 0.0;
   cp->background[2] = 0.0;
   cp->center[0] = 0.0;
   cp->center[1] = 0.0;
   cp->rotate = 0.0;
   cp->pixels_per_unit = 64;
   cp->width = 128;
   cp->height = 128;
   cp->spatial_oversample = 1;
   cp->spatial_filter_radius = 0.5;
   cp->spatial_filter_select = 0 ; //gaussian_filter;
   cp->zoom = 0.0;
   cp->sample_density = 1;
   cp->nbatches = 1;
   cp->ntemporal_samples = 1;
   cp->estimator = 0.0;
   cp->estimator_minimum = 0.0;
   cp->estimator_curve = 0.6;
}

int random_xform ( flam3_genome *g, int excluded )
{
	int ntries = 0;
	while ( ntries++ < 100 )
	{
		int i = random() % g->num_xforms;
		if ( g->xform[i].density > 0.0 && i != excluded )
			return i;
	}
	return -1;
}

void change_colors ( flam3_genome *g, int change_palette )
{
	int i;
	int x0, x1;
	if ( change_palette )
	{
		g->hue_rotation = 0.0;
		g->palette_index = flam3_get_palette ( flam3_palette_random, g->palette, 0.0 );
	}
	for ( i = 0; i < g->num_xforms; i++ )
	{
		g->xform[i].color[0] = flam3_random01();
	}
	x0 = random_xform ( g, -1 );
	x1 = random_xform ( g, x0 );
	if ( x0 >= 0 && ( random() &1 ) ) g->xform[x0].color[0] = 0.0;
	if ( x1 >= 0 && ( random() &1 ) ) g->xform[x1].color[0] = 1.0;
}




void improve_colors(flam3_genome *g, int ntries, int change_palette) {
   int i;
   for (i = 0; i < ntries; i++) {
      change_colors(g, change_palette);
   }
}



flam3_genome* flam3_mutate(flam3_genome* in, flam3_genome* out,
						   double mspeed, int* variations, int nvars)
{
	char notes[10000];
	int count;
	flam3_genome cp_orig, cp_save;

	char *use_vars = getenv("use_vars");
	use_vars = "31,34,35,36,38,43,44,48";
	char *dont_use_vars = getenv("dont_use_vars");
	int seed = 0;
	int bits = 64;
	char default_duv[30]="31,34,35,36,38,43,44,48";

	int rep = 2 ;
	int sym = 0 ;

	int ntries = 10 ;
	double speed = mspeed;
	char action[flam3_max_action_length];
	int i, j;


	memset(&cp_orig, 0, sizeof(flam3_genome));
	memset(&cp_save, 0, sizeof(flam3_genome));

	/* Specify reasonable defaults if nothing is specified */
	if ( !use_vars && !dont_use_vars )
	{
		dont_use_vars = default_duv;
	}


	int ivars[flam3_nvariations];
	int num_ivars = 0;

	if ( use_vars )
	{
		if (nvars > 0)
		{
			memcpy(ivars, variations, nvars*sizeof(int));
			num_ivars = nvars;
		}
		else
		{
// 			ivars[num_ivars++] = 31;
// 			ivars[num_ivars++] = 34;
// 			ivars[num_ivars++] = 35;
// 			ivars[num_ivars++] = 36;
// 			ivars[num_ivars++] = 38;
// 			ivars[num_ivars++] = 43;
// 			ivars[num_ivars++] = 44;
// 			ivars[num_ivars++] = 48;
			/* Set first var to -1 for totally random */
			ivars[0] = -1;
			num_ivars = 0;
		}


		/* Error checking */
		for ( i=0;i<num_ivars;i++ )
		{
			if ( ivars[i]<0 || ivars[i]>=flam3_nvariations )
			{
				exit ( 1 );
			}
		}
	}



	int mutate = 1;
		int did_color;
		notes[0] = 0;
		count = 0;



	flam3_frame frame;
	/* Init random number generators */
	flam3_init_frame(&frame);
	test_cp(&cp_orig);  // just for the width & height
	srandom(seed ? seed : (time(0) + getpid()));

	frame.bits = bits;
	frame.verbose = 0;
	frame.genomes = &cp_orig;
	frame.ngenomes = 1;
	frame.pixel_aspect_ratio = 1.0;
	frame.bytes_per_channel = 1;
	frame.progress = 0;
	frame.nthreads = 1;

	double avg_pix, fraction_black, fraction_white = 0.0;
	double avg_thresh = 20.0;
	double black_thresh = 0.01;
	double white_limit = 0.05;

	unsigned char* image;
	image = (unsigned char *) malloc(3 * cp_orig.width * cp_orig.height);

	do {
		stat_struct stats;

            did_color = 0;
            frame.time = 0.0;

            if (mutate) {

               flam3_genome mutation;
               double r = flam3_random01();

               memset(&mutation, 0, sizeof(flam3_genome));

              flam3_copy(&cp_orig, in);

               if (r < 0.2) {
                  int done = 0;
                  // randomize the variations, usually a large visual effect
                  do {
                     /* Create a random flame, and use the variations */
                     /* to replace those in the original              */
                     flam3_random(&mutation, ivars, num_ivars, sym, cp_orig.num_xforms);
                     for (i = 0; i < cp_orig.num_xforms; i++) {

                        /* Replace if density > 0 or this is the final xform */
                        /*if (cp_orig.xform[i].density > 0.0 || (cp_orig.final_xform_enable==1 && cp_orig.final_xform_index==i)) {*/
                           for (j = 0; j < flam3_nvariations; j++) {
                              if (cp_orig.xform[i].var[j] != mutation.xform[i].var[j]) {
                                 cp_orig.xform[i].var[j] = mutation.xform[i].var[j];

                                 /* Copy parameters for this variation only */
                                 flam3_copy_params(&(cp_orig.xform[i]),&(mutation.xform[i]),j);

                                 done = 1;
                              }
                           }
                        /*}*/
                     }
                  } while (!done);
               } else if (r < 0.4) {
                  // change one xform coefs but not the vars
                  int xf, nxf = 0;
                  flam3_random(&mutation, ivars, num_ivars, sym, 2);
                  nxf = cp_orig.num_xforms;

                  if (0 == nxf) {
					//fprintf(stderr, "no xforms in control point.\n");
                     exit(1);
                  }
                  xf = random()%nxf;

                  // if only two xforms, then change only the translation part
                  if (2 == nxf) {
                     for (j = 0; j < 2; j++)
                        cp_orig.xform[xf].c[2][j] = mutation.xform[0].c[2][j];
                  } else {
                     for (i = 0; i < 3; i++)
                        for (j = 0; j < 2; j++)
                        cp_orig.xform[xf].c[i][j] = mutation.xform[0].c[i][j];
                  }
               } else if (r < 0.5) {
                  flam3_add_symmetry(&cp_orig, 0);
               } else if (r < 0.6) {
                  int b = 1 + random()%6;
                  int same = random()&3;
                  for (i = 0; i < cp_orig.num_xforms; i++) {
                     int copy = (i > 0) && same;
/*                     if (cp_orig.xform[i].density == 0.0) continue;*/
                     if (copy) {
                        for (j = 0; j < 3; j++) {
                           cp_orig.xform[i].post[j][0] = cp_orig.xform[0].post[j][0];
                           cp_orig.xform[i].post[j][1] = cp_orig.xform[0].post[j][1];
                        }
                     } else {
                        if (b&1) {
                           double f = M_PI * flam3_random11();
                           double t[2][2];
                           t[0][0] = (cp_orig.xform[i].c[0][0] * cos(f) +
                              cp_orig.xform[i].c[0][1] * -sin(f));
                           t[0][1] = (cp_orig.xform[i].c[0][0] * sin(f) +
                              cp_orig.xform[i].c[0][1] * cos(f));
                           t[1][0] = (cp_orig.xform[i].c[1][0] * cos(f) +
                              cp_orig.xform[i].c[1][1] * -sin(f));
                           t[1][1] = (cp_orig.xform[i].c[1][0] * sin(f) +
                              cp_orig.xform[i].c[1][1] * cos(f));

                           cp_orig.xform[i].c[0][0] = t[0][0];
                           cp_orig.xform[i].c[0][1] = t[0][1];
                           cp_orig.xform[i].c[1][0] = t[1][0];
                           cp_orig.xform[i].c[1][1] = t[1][1];

                           f *= -1.0;

                           t[0][0] = (cp_orig.xform[i].post[0][0] * cos(f) +
                              cp_orig.xform[i].post[0][1] * -sin(f));
                           t[0][1] = (cp_orig.xform[i].post[0][0] * sin(f) +
                              cp_orig.xform[i].post[0][1] * cos(f));
                           t[1][0] = (cp_orig.xform[i].post[1][0] * cos(f) +
                              cp_orig.xform[i].post[1][1] * -sin(f));
                           t[1][1] = (cp_orig.xform[i].post[1][0] * sin(f) +
                              cp_orig.xform[i].post[1][1] * cos(f));

                           cp_orig.xform[i].post[0][0] = t[0][0];
                           cp_orig.xform[i].post[0][1] = t[0][1];
                           cp_orig.xform[i].post[1][0] = t[1][0];
                           cp_orig.xform[i].post[1][1] = t[1][1];

                        }
                        if (b&2) {
                           double f = 0.2 + flam3_random01();
                           double g;
                           if (random()&1) f = 1.0 / f;
                           if (random()&1) {
                              g = 0.2 + flam3_random01();
                              if (random()&1) g = 1.0 / g;
                           } else
                           g = f;
                           cp_orig.xform[i].c[0][0] /= f;
                           cp_orig.xform[i].c[0][1] /= f;
                           cp_orig.xform[i].c[1][1] /= g;
                           cp_orig.xform[i].c[1][0] /= g;
                           cp_orig.xform[i].post[0][0] *= f;
                           cp_orig.xform[i].post[1][0] *= f;
                           cp_orig.xform[i].post[0][1] *= g;
                           cp_orig.xform[i].post[1][1] *= g;
                        }
                        if (b&4) {
                           double f = flam3_random11();
                           double g = flam3_random11();
                           cp_orig.xform[i].c[2][0] -= f;
                           cp_orig.xform[i].c[2][1] -= g;
                           cp_orig.xform[i].post[2][0] += f;
                           cp_orig.xform[i].post[2][1] += g;
                        }
                     }
                  }
               } else if (r < 0.7) {
                  double s = flam3_random01();
                  did_color = 1;
                  // change just the color
                  if (s < 0.4) {
                     improve_colors(&cp_orig, 100, 0);
                  } else if (s < 0.8) {
                     improve_colors(&cp_orig, 25, 1);
                  } else {
                     cp_orig.palette_index = flam3_get_palette(flam3_palette_random, cp_orig.palette, cp_orig.hue_rotation);
                  }
               } else if (r < 0.8) {
                  int nx = 0;

                  nx = cp_orig.num_xforms;

                  if (nx > 1) {

                     nx = random()%nx;
                     flam3_delete_xform(&cp_orig,nx);

                  }
               } else {
                  int x;
                  flam3_random(&mutation, ivars, num_ivars, sym, cp_orig.num_xforms);

                  // change all the coefs by a little bit
                  for (x = 0; x < cp_orig.num_xforms; x++) {
                        for (i = 0; i < 3; i++) {
                           for (j = 0; j < 2; j++) {
                              cp_orig.xform[x].c[i][j] += speed * mutation.xform[x].c[i][j];
                           /* Eventually, we can mutate the parametric variation coefs here. */
                           }
                        }
                  }
               }

			if (random()&1) {
				double bmin[2], bmax[2];
				flam3_estimate_bounding_box(&cp_orig, 0.01, 100000, bmin, bmax, &frame.rc);
				cp_orig.rot_center[0] = cp_orig.center[0];
				cp_orig.rot_center[1] = cp_orig.center[1];
				cp_orig.center[0] = (bmin[0] + bmax[0]) / 2.0;
				cp_orig.center[1] = (bmin[1] + bmax[1]) / 2.0;
				cp_orig.pixels_per_unit = cp_orig.width / (bmax[0] - bmin[0]);

			}
			if (mutation.num_xforms > 0)
				free(mutation.xform);
		}

		truncate_variations(&cp_orig, 5, action);

		if (!did_color && random()&1) {
			improve_colors(&cp_orig, 100, 0);
		}

		flam3_copy(&cp_save, &cp_orig);
		test_cp(&cp_save);
		frame.genomes = &cp_save;
		flam3_render(&frame, image, cp_save.width, flam3_field_both, 3, 0, &stats);

		{
			int n, tot, totb, totw;
			n = cp_save.width * cp_save.height;
			tot = 0;
			totb = 0;
			totw = 0;
			for (i = 0; i < 3*n; i+=3) {
				tot += (image[i]+image[i+1]+image[i+2]);
				if (0 == image[i] && 0 == image[i+1] && 0 == image[i+2]) totb++;
				if (255 == image[i] && 255 == image[i+1] && 255 == image[i+2]) totw++;
			}

			avg_pix = (tot / (double)(3*n));
			fraction_black = totb / (double)n;
			fraction_white = totw / (double)n;

		}
// 		frame.genomes = &cp_orig;
// 		else {
// 			avg_pix = avg_thresh + 1.0;
// 			fraction_black = black_thresh + 1.0;
// 			fraction_white = white_limit - 1.0;
// 		}

		count++;
	} while ((avg_pix < avg_thresh ||
			fraction_black < black_thresh ||
			fraction_white > white_limit) &&
			count < ntries);

	free(image);

// 	if (ntries == count) {
// 		fprintf(stderr, "warning: reached maximum attempts, giving up.\n");
// 		fflush(stderr);
// 	}
	cp_save.time = ( double ) rep;
	flam3_copy ( out, &cp_orig );

	if (cp_orig.num_xforms > 0)
		free(cp_orig.xform);

	if (cp_save.num_xforms > 0)
		free(cp_save.xform);

	return out;
}




